--[[------------------------------------------------------------------------------------------------

Precision Hands Mod for Farming Simulator 2022
Author: Rusty Tractor Modding

--]]-----------------------------------------------------------------------------------------------


PrecisionHands = {}
PrecisionHands.actionEvents = {}
PrecisionHands.directory = g_currentModDirectory


--Settings Menu Defaults
PrecisionHands.isMetric = true
PrecisionHands.showInHelp = true
PrecisionHands.showInHelpIndex = 2
PrecisionHands.maxDistance = 15
PrecisionHands.maxDistanceIndex = 4
PrecisionHands.pushPullSpeedIndex = 10
PrecisionHands.pushPullSpeed = 1
PrecisionHands.nudgeSpeedIndex = 10
PrecisionHands.nudgeSpeed = 1
PrecisionHands.verticalSpeedIndex = 10
PrecisionHands.verticalSpeed = 1
PrecisionHands.gripStrengthMultiplier = 5
PrecisionHands.gripStrengthIndex = 3
PrecisionHands.rollSpeedIndex = 10
PrecisionHands.rollSpeed = 1
PrecisionHands.superStrengthIndex = 2
PrecisionHands.enableSuperStrength = true

--Variables
PrecisionHands.verticalOffset = 0
PrecisionHands.minDistance = 1
PrecisionHands.state = 0
PrecisionHands.dtInSec = 0
PrecisionHands.pickableObjectInRange = false
PrecisionHands.objectDistance = 0
PrecisionHands.autoRotate = false
PrecisionHands.rotationTarget = {}
PrecisionHands.rotationTarget.x = 0
PrecisionHands.rotationTarget.y = 0
PrecisionHands.rotationTarget.z = 0
PrecisionHands.lastFoundObject = {}
PrecisionHands.playerRiding = false
PrecisionHands.verticalModifierButtonDown = false
PrecisionHands.verticalModifierTime = 0
PrecisionHands.timeTracker = 0
PrecisionHands.keepLevelRotationOffset = 0
PrecisionHands.playerPreviousYPos = 0
PrecisionHands.trigger = false
PrecisionHands.initialised = false

PrecisionHands.superStrength = false
PrecisionHands.lockStrength = false
PrecisionHands.doubleTap = 0
PrecisionHands.doubleTapTime = 0
PrecisionHands.superStrengthValue = 999
PrecisionHands.normalStrengthValue = 0.2
PrecisionHands.superDistanceValue = 12
PrecisionHands.normalDistanceValue = 3
PrecisionHands.superStengthActionId = nil
PrecisionHands.superStengthEventRegistered = false
PrecisionHands.lumberJackLoaded = false


PrecisionHands.setRotationhActionId = nil
PrecisionHands.clearRotationhActionId = nil

--Setup Sounds
PrecisionHands.setTargetSample = createSample("setTargetSound");
loadSample(PrecisionHands.setTargetSample, PrecisionHands.directory.."Sounds/setTarget.ogg", false);
setSamplePitch(PrecisionHands.setTargetSample, 1.3);

--Setup / Configure Mod Settings Folder and Filepath
PrecisionHands.settingsFolder = getUserProfileAppPath() .. "modSettings/"
PrecisionHands.settingsXMLFile = PrecisionHands.settingsFolder .. "PrecisionHands.xml"
createFolder(PrecisionHands.settingsFolder)


--REPLACE Player.MAX_PICKABLE_OBJECT_DISTANCE
function PrecisionHands.playerCheckObjectInRange(self, superFunc)

		if self.maxPickableObjectDistance ~= nil then
			
		end
		return superFunc(self)

end

function PrecisionHands:loadMap(name)
	

	XML.LoadSettings()
	
	if (g_modIsLoaded["FS22_LumberJack"]) then
		--print("LumberJack Is Loaded")
		PrecisionHands.lumberJackLoaded = true
		--LumberJack Mod is in use, do not use superStrength feature of this mod as it will conflict
		PrecisionHands.enableSuperStrength = false
		PrecisionHands.superStrengthIndex = 1
	else
		Player.checkObjectInRange = Utils.overwrittenFunction(Player.checkObjectInRange, PrecisionHands.playerCheckObjectInRange)
		--g_currentMission.player.MAX_PICKABLE_OBJECT_DISTANCE = self.maxPickableObjectDistance
	end
	
	
	if (PrecisionHands.enableSuperStrength == true) then
		Player.registerActionEvents = Utils.appendedFunction(Player.registerActionEvents, PrecisionHands.RegisterSuperStrengthActionEvents)	
	end
	
	--Configure functions to append or override
	FSCareerMissionInfo.saveToXMLFile = Utils.appendedFunction(FSCareerMissionInfo.saveToXMLFile, PrecisionHands.SaveSettings)
	InGameMenuGeneralSettingsFrame.onFrameOpen = Utils.appendedFunction(InGameMenuGeneralSettingsFrame.onFrameOpen, GUI.initGui)
	Player.pickUpObject = Utils.overwrittenFunction(Player.pickUpObject, PrecisionHands.pickUpObject)
	Player.pickUpObject = Utils.appendedFunction(Player.pickUpObject, PrecisionHands.PickupStarted)
	Player.onInputRotateObjectVertically = Utils.appendedFunction(Player.onInputRotateObjectVertically, PrecisionHands.manualRotateObjectStarted)
	Player.onInputRotateObjectHorizontally = Utils.appendedFunction(Player.onInputRotateObjectHorizontally, PrecisionHands.manualRotateObjectStarted)
	Player.removeActionEvents = Utils.appendedFunction(Player.removeActionEvents, PrecisionHands.RemoveActionEvents)
	

end




function PrecisionHands:update(dt)

	-- Dedicated Server has no player
	if g_currentMission.player==nil then
		return
	end
		
	-- CHANGE GLOBAL VALUES ON FIRST RUN
	if (g_gameStateManager:getGameState()==GameState.PLAY and PrecisionHands.initialised==false) then
		g_currentMission.player.maxPickableMass = PrecisionHands.normalStrengthValue
		g_currentMission.player.maxPickableObjectDistance = PrecisionHands.normalDistanceValue
		PrecisionHands.superStrength = false
		PrecisionHands.lockStrength = false
		PrecisionHands.initialised = true

	end
	
	--Rotate to defined position if AutoRotate enabled
	if g_currentMission.player.isCarryingObject and PrecisionHands.autoRotate then
		local node, nodeChild = g_currentMission.player.model:getKinematicHelpers()
		setWorldRotation(nodeChild, PrecisionHands.rotationTarget.x, PrecisionHands.rotationTarget.y, PrecisionHands.rotationTarget.z)
		g_currentMission.player:rotateObject(0,0,0,0)
	end
		

	if g_currentMission.player.pickedUpObject ~= nil then

		local x, y, z = getWorldTranslation(g_currentMission.player.rootNode)
		local verticalVelocity = (y - PrecisionHands.playerPreviousYPos)/dt
		if verticalVelocity > 0.005 then
			PrecisionHands.CheckIfRiding()
		end
		PrecisionHands.playerPreviousYPos = y

		PrecisionHands.dtInSec = dt * 0.001
		PrecisionHands.timeTracker = PrecisionHands.timeTracker + PrecisionHands.dtInSec

		-- Determine if the Vertical Modifier Key is down
		PrecisionHands.verticalModifierButtonDown = false
		if g_inputBinding.nameActions.VERTICAL_MODIFIER.bindings ~= nil then
			for i=1,#g_inputBinding.nameActions.VERTICAL_MODIFIER.bindings do
				if g_inputBinding.nameActions.VERTICAL_MODIFIER.bindings[i].hasFrameTriggered then
					PrecisionHands.verticalModifierButtonDown = true
				end
			end
		end
	end

	-- HANDLE AUTO-ROTATE CONTROLS
	if g_currentMission.player.isObjectInRange == true and g_currentMission.player.isCarryingObject == false then
		PrecisionHands:RegisterPickupActionEvents()
	elseif g_currentMission.player.isObjectInRange == false and PrecisionHands.clearRotationhActionId == nil and g_currentMission.player.isCarryingObject == false and PrecisionHands.autoRotate then
		PrecisionHands:RegisterClearAutorotateTargetActionEvent()
	elseif g_currentMission.player.isObjectInRange == false and PrecisionHands.clearRotationhActionId == nil and g_currentMission.player.isCarryingObject == false then
		PrecisionHands.RemoveActionEvents()
	end

	
			
	if g_currentMission.player.isEntered and not g_gui:getIsGuiVisible() and PrecisionHands.enableSuperStrength == true then
	
			-- DETECT DOUBLE TAP SUPER STRENGTH KEY
			if PrecisionHands.doubleTap ~= 0 then
				if PrecisionHands.doubleTap == 1 then
					PrecisionHands.doubleTapTime = PrecisionHands.doubleTapTime + dt
					if PrecisionHands.doubleTapTime > 500 then
						PrecisionHands.doubleTap = 0
						PrecisionHands.doubleTapTime = 0
					end
				else
					--print("DOUBLE TAP")
					PrecisionHands.lockStrength = not PrecisionHands.lockStrength
					PrecisionHands.doubleTap = 0
					PrecisionHands.doubleTapTime = 0
				end
			end

			-- CHANGE COLOUR OF THE CURSOR/HAND ICON
			if PrecisionHands.superStrength then
				-- Make hand BRIGHTER when super strength is ON
				g_currentMission.player.aimOverlay:setColor(1, 1, 1, 1.0)
				g_currentMission.player.pickedUpObjectOverlay:setColor(1, 1, 1, 1.0)
			else
				if g_currentMission.player.isObjectInRange and g_currentMission.player.lastFoundObjectMass > g_currentMission.player.maxPickableMass then
					-- Make hand RED when objects are too heavy to pick up
					g_currentMission.player.pickedUpObjectOverlay:setColor(1.0, 0.1, 0.1, 0.5)
				else
					-- Make cursor/hand GREY for everything else
					g_currentMission.player.pickedUpObjectOverlay:setColor(1, 1, 1, 0.3)
				end
				g_currentMission.player.aimOverlay:setColor(1, 1, 1, 0.3)
			end
	end
		
	if (PrecisionHands.superStengthEventRegistered == true and PrecisionHands.enableSuperStrength == false) and g_currentMission.inGameMenu.isOpen == false then
		g_inputBinding:removeActionEvent(PrecisionHands.superStengthActionId)
		PrecisionHands.superStengthEventRegistered = false
		PrecisionHands.doubleTap = 0
		PrecisionHands.lockStrength = false
		PrecisionHands.toggleStrength("none", 0)
		g_currentMission.player.pickedUpObjectOverlay:setColor(1, 1, 1, 0.3)
		g_currentMission.player.aimOverlay:setColor(1, 1, 1, 0.3)
		g_currentMission.player.maxPickableMass = PrecisionHands.normalStrengthValue
		g_currentMission.player.maxPickableObjectDistance = PrecisionHands.normalDistanceValue
		PrecisionHands.lockStrength = false
	end

	if (PrecisionHands.enableSuperStrength == true and g_currentMission.inGameMenu.isOpen == false and PrecisionHands.superStengthEventRegistered == false) then
		PrecisionHands.RegisterSuperStrengthActionEvents()
		PrecisionHands.superStengthEventRegistered = true
	end


end




-- EVENTS SYSTEM --------------------------------------------------------------------------------
function PrecisionHands:RegisterSuperStrengthActionEvents()

	if g_client then

		for _, displayActionEvent in ipairs(g_inputBinding:getDisplayActionEvents()) do
			if displayActionEvent.action.name == "PRECISIONHANDS_SUPERSTRENGTH" then
					return
			end
		end
		
		local _, actionEventId = g_inputBinding:registerActionEvent('PRECISIONHANDS_SUPERSTRENGTH', g_currentMission.player, PrecisionHands.toggleStrength, true, true, false, true)
		g_inputBinding:setActionEventTextPriority(actionEventId, GS_PRIO_VERY_LOW)
		g_inputBinding:setActionEventActive(actionEventId, true)
		g_inputBinding:setActionEventText(actionEventId, g_i18n:getText("menu_TOGGLE_STRENGTH"))

		PrecisionHands.superStengthActionId = actionEventId
		
		g_currentMission.player.inputInformation.registrationList[InputAction.PRECISIONHANDS_SUPERSTRENGTH] = {
			text = g_i18n:getText("menu_TOGGLE_STRENGTH"),
			triggerAlways = false,
			triggerDown = true,
			eventId = actionEventId,
			textVisibility = true,
			triggerUp = true,
			callback = PrecisionHands.toggleStrength,
			activeType = Player.INPUT_ACTIVE_TYPE.IS_MOVEMENT
		}

		PrecisionHands.superStengthEventRegistered = true
	end
	
end


function PrecisionHands:RegisterDropActionEvents()


		


		PrecisionHands.RemoveActionEvents()
		
		--Register Action Events
		result, PrecisionHands.AutoRotateEvent = g_inputBinding:registerActionEvent('AUTO_ROTATE',g_currentMission.player, PrecisionHands.AutoRotateToggle ,false ,true ,false ,true)
		g_inputBinding:setActionEventTextPriority(PrecisionHands.AutoRotateEvent, GS_PRIO_VERYHIGH);
		PrecisionHands:AddToActionEventArray(PrecisionHands.AutoRotateEvent)		
		if PrecisionHands.autoRotate then
			g_inputBinding:setActionEventText(PrecisionHands.AutoRotateEvent, g_i18n:getText('input_AUTO_ROTATE') .. " (ENABLED)")
		else
			g_inputBinding:setActionEventText(PrecisionHands.AutoRotateEvent, g_i18n:getText('input_AUTO_ROTATE') .. " (DISABLED)")
		end
		
		result, PrecisionHands.pushActionEvent = g_inputBinding:registerActionEvent('PUSH_FURTHER',g_currentMission.player, PrecisionHands.PushFurther ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.pushActionEvent)
		
		result, PrecisionHands.pullActionEvent = g_inputBinding:registerActionEvent('PULL_CLOSER',g_currentMission.player, PrecisionHands.PullCloser ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.pullActionEvent)
		
		result, PrecisionHands.NudgeLeftEvent = g_inputBinding:registerActionEvent('NUDGE_LEFT',g_currentMission.player, PrecisionHands.NudgeLeft ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.NudgeLeftEvent)
		
		result, PrecisionHands.NudgeRightEvent = g_inputBinding:registerActionEvent('NUDGE_RIGHT',g_currentMission.player, PrecisionHands.NudgeRight ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.NudgeRightEvent)
		
		result, PrecisionHands.NudgeFurtherEvent = g_inputBinding:registerActionEvent('NUDGE_FURTHER',g_currentMission.player, PrecisionHands.NudgeFurther ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.NudgeFurtherEvent)

		result, PrecisionHands.NudgeCloserEvent = g_inputBinding:registerActionEvent('NUDGE_CLOSER',g_currentMission.player, PrecisionHands.NudgeCloser ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.NudgeCloserEvent)
		
		result, PrecisionHands.VerticalModifierEventDown = g_inputBinding:registerActionEvent('VERTICAL_MODIFIER',g_currentMission.player, PrecisionHands.VerticalModifier ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.VerticalModifierEventDown)		
		g_inputBinding:setActionEventTextPriority(PrecisionHands.VerticalModifierEventDown, GS_PRIO_VERYHIGH);
		
		result, PrecisionHands.MoveUpVerticallyEvent = g_inputBinding:registerActionEvent('MOVE_UP_VERTICALLY',g_currentMission.player, PrecisionHands.MoveUpVertically ,true ,false ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.MoveUpVerticallyEvent)
		
		result, PrecisionHands.MoveDownVerticallyEvent = g_inputBinding:registerActionEvent('MOVE_DOWN_VERTICALLY',g_currentMission.player, PrecisionHands.MoveDownVertically ,true ,false ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.MoveDownVerticallyEvent)
		
		result, PrecisionHands.RollLeftEvent = g_inputBinding:registerActionEvent('ROLL_LEFT',g_currentMission.player, PrecisionHands.RollLeft ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.RollLeftEvent)

		result, PrecisionHands.RollRightEvent = g_inputBinding:registerActionEvent('ROLL_RIGHT',g_currentMission.player, PrecisionHands.RollRight ,false ,true ,true ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.RollRightEvent)

		for _, actionEvent in ipairs(PrecisionHands.actionEvents) do
			g_inputBinding:setActionEventTextVisibility(actionEvent, PrecisionHands.showInHelp)
			g_inputBinding:setActionEventActive(actionEvent, true)
		end
		
		g_inputBinding:setActionEventTextVisibility(PrecisionHands.AutoRotateEvent, PrecisionHands.showInHelp)
		
	
end


function PrecisionHands:RegisterPickupActionEvents()


		PrecisionHands.RemoveActionEvents()
		
		--Register Action Events
		result, PrecisionHands.SetAutoRotateTargetEvent = g_inputBinding:registerActionEvent('SET_AUTO_ROTATE_TARGET',g_currentMission.player, PrecisionHands.SetAutoRotateTarget ,false ,true ,false ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.SetAutoRotateTargetEvent)	
		
		PrecisionHands.setRotationhActionId = PrecisionHands.SetAutoRotateTargetEvent
		
		for _, actionEvent in ipairs(PrecisionHands.actionEvents) do
			g_inputBinding:setActionEventTextVisibility(actionEvent, PrecisionHands.showInHelp)
		end

end

function PrecisionHands:RegisterClearAutorotateTargetActionEvent()

		PrecisionHands.RemoveActionEvents()
		
		--Register Action Events
		result, PrecisionHands.ClearAutorotateTargetActionEvent = g_inputBinding:registerActionEvent('CLEAR_AUTO_ROTATE_TARGET',g_currentMission.player, PrecisionHands.ClearAutoRotateTarget ,false ,true ,false ,true)
		PrecisionHands:AddToActionEventArray(PrecisionHands.ClearAutorotateTargetActionEvent)	
		
		PrecisionHands.clearRotationhActionId = PrecisionHands.ClearAutorotateTargetActionEvent
		
		for _, actionEvent in ipairs(PrecisionHands.actionEvents) do
			g_inputBinding:setActionEventTextVisibility(actionEvent, PrecisionHands.showInHelp)
		end

end
	

function PrecisionHands.VerticalModifier(actionEvent)
--Not used
end

function PrecisionHands:AddToActionEventArray(actionEvent)
	table.insert(PrecisionHands.actionEvents, actionEvent)
end

function PrecisionHands:RemoveActionEvents()

	if g_client then
		for _, actionEvent in ipairs(PrecisionHands.actionEvents) do
			g_inputBinding:removeActionEvent(actionEvent)
		end
		PrecisionHands.actionEvents = {}
		PrecisionHands.setRotationhActionId = nil
		PrecisionHands.clearRotationhActionId = nil
	end	
		
end



-- ACTION FUNCTIONS -----------------------------------------------------------

function PrecisionHands:SetAutoRotateTarget()
	if g_currentMission.player.lastFoundAnyObject ~= nil then
		PrecisionHands.rotationTarget.x,PrecisionHands.rotationTarget.y,PrecisionHands.rotationTarget.z = getWorldRotation(g_currentMission.player.lastFoundAnyObject)
		PrecisionHands.autoRotate = true
		g_inputBinding:setActionEventText(PrecisionHands.AutoRotateEvent, g_i18n:getText('input_AUTO_ROTATE') .. " (ENABLED)")
		playSample(PrecisionHands.setTargetSample,1,0.6,0,0,0)
	end
end

function PrecisionHands:ClearAutoRotateTarget()
		PrecisionHands.autoRotate = false
		PrecisionHands.RemoveActionEvents()
end


function PrecisionHands:AutoRotateToggle()

		if PrecisionHands.autoRotate == false then
			PrecisionHands.autoRotate = true
			g_inputBinding:setActionEventText(PrecisionHands.AutoRotateEvent, g_i18n:getText('input_AUTO_ROTATE') .. " (ENABLED)")
			playSample(PrecisionHands.setTargetSample,1,0.6,0,0,0)

		else
			PrecisionHands.autoRotate = false
			g_inputBinding:setActionEventText(PrecisionHands.AutoRotateEvent, g_i18n:getText('input_AUTO_ROTATE') .. " (DISABLED)")
		end
end

function PrecisionHands:RollLeft()
	PrecisionHands.autoRotate = false
	if PrecisionHands.InputDelayCheck then 
		g_currentMission.player:rotateObject((PrecisionHands.rollSpeed*0.5), 0.0, 0.0, 1.0)
	end
end

function PrecisionHands:RollRight()
	PrecisionHands.autoRotate = false
	if PrecisionHands.InputDelayCheck then 
		g_currentMission.player:rotateObject((PrecisionHands.rollSpeed*0.5), 0.0, 0.0, -1.0)
	end
end


function PrecisionHands:PerformNudge(axisInputVal,dx,dy,dz)

	local offsetAmount = (0.00625 * PrecisionHands.nudgeSpeed)*axisInputVal
	PrecisionHands:UpdateDistance()
	local dx,dy,dz = localDirectionToWorld(g_currentMission.player.cameraNode,dx*offsetAmount,dy*offsetAmount,dz*offsetAmount)
	local wx, wy, wz = getWorldTranslation(g_currentMission.player.model:getKinematicHelpers())
	setWorldTranslation(kinematicHelperNode,wx +dx, wy+dy, wz+dz)
end

function PrecisionHands:NudgeLeft(_,axisInputVal)
	if PrecisionHands.InputDelayCheck then 
		PrecisionHands:PerformNudge(axisInputVal,-1,0,0)
	end
end

function PrecisionHands:NudgeRight(_,axisInputVal)
	if PrecisionHands.InputDelayCheck then 
		PrecisionHands:PerformNudge(axisInputVal,1,0,0)
	end
end

function PrecisionHands:NudgeFurther(_,axisInputVal)

	if PrecisionHands.InputDelayCheck then 
		PrecisionHands:PerformNudge(axisInputVal,0,0,-1)
	end
end

function PrecisionHands:NudgeCloser(_,axisInputVal)
	if PrecisionHands.InputDelayCheck then 
		PrecisionHands:PerformNudge(axisInputVal,0,0,1)
	end
end

function PrecisionHands:MoveUpVertically(_,axisInputVal)

	if PrecisionHands.InputDelayCheck then 

			if g_currentMission.player.isCarryingObject == true then
				local wx, wy, wz = getWorldTranslation(g_currentMission.player.model:getKinematicHelpers())
				local offsetAmount = 0.05 * (PrecisionHands.verticalSpeed * axisInputVal)
				PrecisionHands.CheckIfRiding()
				setWorldTranslation(kinematicHelperNode,wx, wy+offsetAmount, wz)
				PrecisionHands.verticalOffset = PrecisionHands.verticalOffset + offsetAmount
			end
			
	end

end

function PrecisionHands:MoveDownVertically(_,axisInputVal)

	if PrecisionHands.InputDelayCheck then 
	
		if g_currentMission.player.isCarryingObject == true then
			local wx, wy, wz = getWorldTranslation(g_currentMission.player.model:getKinematicHelpers())
			local offsetAmount = 0.05 *(PrecisionHands.verticalSpeed * axisInputVal)
			PrecisionHands.CheckIfRiding()
			setWorldTranslation(kinematicHelperNode,wx, wy - offsetAmount , wz)
			PrecisionHands.verticalOffset = PrecisionHands.verticalOffset - offsetAmount
		end
		
	end
	
end


function PrecisionHands:PullCloser(_,axisInputVal)

	if PrecisionHands.InputDelayCheck then 

		--Function to move object closer to player
		if g_currentMission.player.isCarryingObject == true then

			local currentInputHelpMode = g_currentMission.inputManager:getInputHelpMode()	
			local dx,dy,dz = 0,0,0
			
			local offsetAmount = -0.125 * (PrecisionHands.pushPullSpeed * axisInputVal)
			dx,dy,dz = localDirectionToWorld(self.cameraNode, 0,0,offsetAmount )
			local wx, wy, wz = getWorldTranslation(self.model:getKinematicHelpers())
				
				if PrecisionHands.verticalModifierButtonDown then
					PrecisionHands:MoveDownVertically(_,axisInputVal)
				else
					
					PrecisionHands:UpdateDistance()
					if PrecisionHands.objectDistance > PrecisionHands.minDistance then
						setWorldTranslation(kinematicHelperNode,wx -dx, wy-dy, wz-dz)
					end
				end
		else
			return
		end
		
	end	

end


function PrecisionHands:PushFurther(param1,axisInputVal)


	if PrecisionHands.InputDelayCheck then  
	
			--Function to move object further from the player
			if g_currentMission.player.isCarryingObject == true  then
				
				local currentInputHelpMode = g_currentMission.inputManager:getInputHelpMode()	
				local dx,dy,dz = 0,0,0
				local offsetAmount = -0.125 * (PrecisionHands.pushPullSpeed * axisInputVal)

				dx,dy,dz = localDirectionToWorld(self.cameraNode, 0,0,offsetAmount )
				local wx, wy, wz = getWorldTranslation(self.model:getKinematicHelpers())


					if PrecisionHands.verticalModifierButtonDown then
						PrecisionHands:MoveUpVertically(_,axisInputVal)
					else
				
				
					PrecisionHands:UpdateDistance()
					
						if PrecisionHands.objectDistance < PrecisionHands.maxDistance then
							setWorldTranslation(kinematicHelperNode,wx +dx, wy+dy, wz+dz)
						end
					end
			else
				return
			end
	end
end


function PrecisionHands:pickUpObject(thePlayer, state, noEventSend) --Overwritten Function

	PlayerPickUpObjectEvent.sendEvent(self, state, noEventSend)
	

    if self.isServer then

        if state and (self.isObjectInRange and self.lastFoundObject ~= nil and entityExists(self.lastFoundObject)) and not self.isCarryingObject then
            local constr = JointConstructor.new()

            -- disable collision with CCT
            self.pickedUpObjectCollisionMask = getCollisionMask(self.lastFoundObject)
            local newPickedUpObjectCollisionFlag = bitXOR(bitAND(self.pickedUpObjectCollisionMask, CollisionMask.PLAYER_MOVEMENT), self.pickedUpObjectCollisionMask)
            setCollisionMask(self.lastFoundObject, newPickedUpObjectCollisionFlag)
			
	
			PrecisionHands.lastFoundObject = self.lastFoundObject --EXTRA CODE
			local kinematicHelperNode, kinematicHelperNodeChild = self.model:getKinematicHelpers()
			
            constr:setActors(kinematicHelperNode, self.lastFoundObject)


            for i=0, 2 do
                constr:setRotationLimit(i, 0, 0)
                constr:setTranslationLimit(i, true, 0, 0)
            end

            -- set position of joint to center of the object
            local mx, my, mz = getCenterOfMass(self.lastFoundObject)
            local wx, wy, wz = localToWorld(self.lastFoundObject, mx, my, mz)
            constr:setJointWorldPositions(wx, wy, wz, wx, wy, wz)

            local nx, ny, nz = localDirectionToWorld(self.lastFoundObject, 1, 0, 0)
            constr:setJointWorldAxes(nx, ny, nz, nx, ny, nz)

            local yx, yy, yz = localDirectionToWorld(self.lastFoundObject, 0, 1, 0)
            constr:setJointWorldNormals(yx, yy, yz, yx, yy, yz)
            constr:setEnableCollision(false)

            -- Update child, used for object rotation by player
            setWorldTranslation(kinematicHelperNodeChild, wx, wy, wz)
            setWorldRotation(kinematicHelperNodeChild, getWorldRotation(self.lastFoundObject))

            -- set spring/damper ?!
            local dampingRatio = 0.8
            local mass = getMass(self.lastFoundObject)

            local rotationLimitSpring = {}
            local rotationLimitDamper = {}
            for i=1, 3 do
                rotationLimitSpring[i] = mass * 60
                rotationLimitDamper[i] = dampingRatio * 2 * math.sqrt( mass * rotationLimitSpring[i] )
         
            end
            constr:setRotationLimitSpring(rotationLimitSpring[1], rotationLimitDamper[1], rotationLimitSpring[2], rotationLimitDamper[2], rotationLimitSpring[3], rotationLimitDamper[3])

            local translationLimitSpring = {}
            local translationLimitDamper = {}
            for i=1, 3 do
                translationLimitSpring[i] = mass * 60
                translationLimitDamper[i] = dampingRatio * 2 * math.sqrt( mass * translationLimitSpring[i] )
   
            end
            constr:setTranslationLimitSpring(translationLimitSpring[1], translationLimitDamper[1], translationLimitSpring[2], translationLimitDamper[2], translationLimitSpring[3], translationLimitDamper[3])

            local forceAcceleration = 4
            local forceLimit = forceAcceleration * mass * 40.0
			forceLimit = forceLimit * PrecisionHands.gripStrengthMultiplier -- EXTRA CODE --
            constr:setBreakable(forceLimit, forceLimit)

            self.pickedUpObjectJointId = constr:finalize()

            addJointBreakReport(self.pickedUpObjectJointId, "onPickedUpObjectJointBreak", self)

            self.pickedUpObject = self.lastFoundObject
            self.isCarryingObject = true
            Player.PICKED_UP_OBJECTS[self.pickedUpObject] = true

            local object = g_currentMission:getNodeObject(self.pickedUpObject)
            if object ~= nil then
                object.thrownFromPosition = nil
            end
			
			
        else
            if self.pickedUpObjectJointId ~= nil then
                removeJoint(self.pickedUpObjectJointId)
                self.pickedUpObjectJointId = nil
                self.isCarryingObject = false
                Player.PICKED_UP_OBJECTS[self.pickedUpObject] = false

                if entityExists(self.pickedUpObject) then
                    local vx, vy, vz = getLinearVelocity(self.pickedUpObject)
                    if vx ~= nil then -- in case the object has switched from dynamic to kinematic
                        vx = MathUtil.clamp(vx, -5, 5)
                        vy = MathUtil.clamp(vy, -5, 5)
                        vz = MathUtil.clamp(vz, -5, 5)
                        setLinearVelocity(self.pickedUpObject, vx, vy, vz)
                    end
                    setCollisionMask(self.pickedUpObject, self.pickedUpObjectCollisionMask)
                    self.pickedUpObjectCollisionMask = 0
		
                end

                local object = g_currentMission:getNodeObject(self.pickedUpObject)
                if object ~= nil then
                    object.thrownFromPosition = nil
                end
                self.pickedUpObject = nil
				PrecisionHands.lastFoundObject = nil
            end
        end
    end
end


function PrecisionHands:manualRotateObjectStarted(_, inputValue)
	if g_client then
		if (inputValue ~= 0) and PrecisionHands.autoRotate then
			--print(inputValue)
			PrecisionHands.autoRotate = false
			g_inputBinding:setActionEventText(PrecisionHands.AutoRotateEvent, g_i18n:getText('input_AUTO_ROTATE') .. " (DISABLED)")
		end
	end

end





-- STANDARD FUNCTIONS ----------------------------------------------------

function PrecisionHands:SaveSettings()
	XML.SaveSettings()
end
function PrecisionHands:LoadSettings()
	XML.LoadSettings()
end


function PrecisionHands:CheckIfRiding()
	--Check if player is on top of held object
	local x, y, z = getWorldTranslation(g_currentMission.player.rootNode)
	raycastClosest(x,y,z, 0, -1, 0, "CheckIfRiding_Callback", 3, PrecisionHands, CollisionMask.PLAYER_MOVEMENT)

end

function PrecisionHands:CheckIfRiding_Callback(hitObjectId, x, y, z, distance)
	
	local rootVehicle = g_currentMission:getNodeObject(g_currentMission.player.pickedUpObject)
	if rootVehicle ~= nil then
		local rootVehicleId = g_currentMission.player.pickedUpObject
		if rootVehicle.findRootVehicle ~= nil then
			rootVehicleId = rootVehicle:findRootVehicle().rootNode
		end
		
		local hitObject = g_currentMission:getNodeObject(hitObjectId)
		local hitObjectRootId = hitObjectId
		if hitObject ~= nil then
			if hitObject.findRootVehicle ~= nil then
				hitObjectRootId = hitObject:findRootVehicle().rootNode
			end
		end
		
		
		if hitObjectRootId == rootVehicleId then
			PrecisionHands.playerRiding = true
			g_currentMission.player.playerStateMachine:activateState("drop")
		else
			PrecisionHands.playerRiding = false
		end
	end
	
end


function PrecisionHands:InputDelayCheck()
	if PrecisionHands.timeTracker >= (0.016666666666*4) then 
		PrecisionHands.timeTracker = 0
			return true
		else
			return false
	end
end

function PrecisionHands:UpdateDistance()

		local targetCamX, targetCamY, targetCamZ = getWorldTranslation(g_currentMission.player.cameraNode);
		local nodeX, nodeY, nodeZ = getWorldTranslation(kinematicHelperNodeChild)
        local dirX, dirY, dirZ = targetCamX-nodeX, targetCamY-nodeY, targetCamZ-nodeZ
        PrecisionHands.objectDistance = MathUtil.vector3Length(dirX, dirY, dirZ)

end

function PrecisionHands:SetDistance()
		
		local targetCamX, targetCamY, targetCamZ = getWorldTranslation(g_currentMission.player.cameraNode);
		local nodeX, nodeY, nodeZ = g_currentMission.player.lastFoundObjectHitPoint[1],g_currentMission.player.lastFoundObjectHitPoint[2],g_currentMission.player.lastFoundObjectHitPoint[3]
        local dirX, dirY, dirZ = targetCamX-nodeX, targetCamY-nodeY, targetCamZ-nodeZ
        PrecisionHands.minDistance = MathUtil.vector3Length(dirX, dirY, dirZ)
		
end


function PrecisionHands:ActivatePickupActionEvents()
		for _, actionEvent in ipairs(PrecisionHands.actionEvents) do
			
		end

end



function PrecisionHands:PickupStarted(state, noEventSend)


	if g_currentMission.player.rootNode == self.rootNode then
		self.pickedUpObject = self.lastFoundObject

		if g_currentMission.player.pickedUpObject ~= nil then
			PrecisionHands.RegisterDropActionEvents()
			kinematicHelperNode, kinematicHelperNodeChild = g_currentMission.player.model:getKinematicHelpers()
			_,PrecisionHands.startingY,_ = getRotation(kinematicHelperNodeChild)
			PrecisionHands.state = 2
			PrecisionHands.verticalOffset = 0
			PrecisionHands.UpdateDistance()


		else
			if PrecisionHands.state ~= 0 then
				PrecisionHands.state = 0
				PrecisionHands:RegisterPickupActionEvents()
				local kinematicHelperNode, kinematicHelperNodeChild = g_currentMission.player.model:getKinematicHelpers()
				local x,y,z = localToWorld(g_currentMission.player.cameraNode, 0,0,-1.0)
				setWorldTranslation(kinematicHelperNode, x, y, z) -- EXTRA CODE
				
			end
		end
	end
	
end

function PrecisionHands:ResetVerticalOffset()
	
		local reverseOffset = PrecisionHands.verticalOffset * -1
		local _, objectTransform = g_currentMission.player.model:getKinematicHelpers()
		local wx, wy, wz = getWorldTranslation(objectTransform)
		setWorldTranslation(objectTransform,wx, wy + reverseOffset, wz)
		PrecisionHands.verticalOffset = 0

end


function PrecisionHands:toggleStrength(name, state)
	
	if g_currentMission.player.isEntered and not g_gui:getIsGuiVisible() then

		if state == 1 then
			PrecisionHands.doubleTap = PrecisionHands.doubleTap + 1
		else

		end
		
		if PrecisionHands.lockStrength then
			--print("SUPER STRENGTH LOCKED")
				PrecisionHands.superStrength = true
				g_currentMission.player.superStrengthEnabled = PrecisionHands.superStrength
				--g_currentMission.player.MAX_PICKABLE_OBJECT_MASS = PrecisionHands.superStrengthValue
				g_currentMission.player.maxPickableMass = PrecisionHands.superStrengthValue
				--g_currentMission.player.MAX_PICKABLE_OBJECT_DISTANCE = PrecisionHands.superDistanceValue
				g_currentMission.player.maxPickableObjectDistance = PrecisionHands.superDistanceValue
				--print(g_currentMission.player.maxPickableObjectDistance)
			
		else
			if state == 0 then
				--print("SUPER STRENGTH OFF")
				PrecisionHands.superStrength = false
				g_currentMission.player.maxPickableMass = PrecisionHands.normalStrengthValue
				g_currentMission.player.maxPickableObjectDistance = PrecisionHands.normalDistanceValue
				g_currentMission.player.superStrengthEnabled = PrecisionHands.superStrength
				--print(g_currentMission.player.maxPickableObjectDistance)
			end
			
			if state == 1 then
				--print("SUPER STRENGTH ON")
				PrecisionHands.superStrength = true
				g_currentMission.player.maxPickableMass = PrecisionHands.superStrengthValue
				g_currentMission.player.maxPickableObjectDistance = PrecisionHands.superDistanceValue
				g_currentMission.player.superStrengthEnabled = PrecisionHands.superStrength
				--print(g_currentMission.player.maxPickableObjectDistance)
				
			end

		end
		
	end
end


---------------------------------------
addModEventListener(PrecisionHands)


